from asyncio import Lock
from collections import deque
from io import BytesIO
from typing import Deque, Optional, TYPE_CHECKING, Tuple

from aio_pika import RobustQueue
from aio_pika.abc import AbstractIncomingMessage, ConsumerTag, HeadersType
from no_exception import NoException
from pamqp.common import Arguments

from .base_device import RabbitMQBaseInputDevice
from .transaction import BaseTransaction, EmptyTransaction, RabbitMQIncomingMessageTransaction

if TYPE_CHECKING:
    from aio_rabbitmq_utils import RabbitMQConsumeInputDeviceManager


class RabbitMQInputConsumeDevice(RabbitMQBaseInputDevice):
    def __init__(
        self,
        device_manager: "RabbitMQConsumeInputDeviceManager",
        device_name: str,
        use_transaction: bool,
        consumer_arguments: Arguments = None,
    ):
        self._device_manager = device_manager
        self._device_name = device_name
        self._use_transaction = use_transaction
        self._consumer_arguments = consumer_arguments

        self._lock = Lock()
        self._queue: Optional[RobustQueue] = None
        self._consumer_tag: Optional[ConsumerTag] = None
        self._inner_queue: Deque[Tuple[BytesIO, HeadersType, BaseTransaction]] = deque([])
        self._max_delivery_tag: Optional[int] = None

    async def _inner_commit_rollback_all_messages(self, commit: bool) -> None:
        if not self._max_delivery_tag:
            raise Exception("No messages to commit/rollback")

        channel = await self._device_manager.channel
        aiormq_channel = await channel.get_underlay_channel()
        # delivery_tag=0 + multiple=True means ack/nack all un-acked messages in the channel
        func = aiormq_channel.basic_ack if commit else aiormq_channel.basic_nack
        return await func(delivery_tag=self._max_delivery_tag, multiple=True)

    async def commit_all_messages(self) -> None:
        return await self._inner_commit_rollback_all_messages(commit=True)

    async def rollback_all_messages(self) -> None:
        return await self._inner_commit_rollback_all_messages(commit=False)

    @property
    def has_messages_available(self) -> bool:
        return len(self._inner_queue) > 0

    @property
    async def queue(self) -> RobustQueue:
        if self._queue is None or self._queue.channel.is_closed:
            channel = await self._device_manager.channel
            self._queue = await channel.get_queue(self._device_name)
        return self._queue

    async def _inner_consume(
        self,
        incoming_message: AbstractIncomingMessage,
    ) -> None:
        transaction = RabbitMQIncomingMessageTransaction(incoming_message) \
            if self._use_transaction else EmptyTransaction()
        async with self._lock:
            self._inner_queue.append(
                (
                    BytesIO(incoming_message.body),
                    incoming_message.headers,
                    transaction,
                ),
            )

    async def read(
        self,
    ) -> Optional[Tuple[BytesIO, HeadersType, BaseTransaction]]:
        async with self._lock:
            try:
                data, headers, transaction = self._inner_queue.popleft()
                if self._use_transaction:
                    self._max_delivery_tag = max(
                        transaction._incoming_message.delivery_tag,
                        self._max_delivery_tag or 0,
                    )
                return data, headers, transaction
            except IndexError:
                return None

    async def connect(self) -> None:
        self._consumer_tag = await (await self.queue).consume(
            self._inner_consume,
            no_ack=not self._use_transaction,
            arguments=self._consumer_arguments,
        )

    async def close(self) -> None:
        async with self._lock:
            for _, _, transaction in self._inner_queue:
                with NoException():
                    await transaction.rollback()
            self._inner_queue.clear()
        await (await self.queue).cancel(self._consumer_tag)



import random
from abc import ABC
from typing import Dict, List, Optional

from aio_pika import connect_robust
from aio_pika.abc import AbstractRobustChannel, AbstractRobustConnection
from aio_pika.connection import make_url
from aiormq.connection import DEFAULT_PORTS
from no_exception import NoException
from pamqp.constants import DEFAULT_PORT

from .base_device_manager import RabbitMQBaseDeviceManager


class RabbitMQDeviceManager(RabbitMQBaseDeviceManager, ABC):
    def __init__(
        self,
        hosts: List[str],
        user: str,
        password: str,
        vhost: str,
        publisher_confirms: bool,
        channel_qos_kwargs: Dict[str, int | float | bool | None] = None,
        use_transaction: bool = False,
        use_ssl: bool = True,
        port: int = None,
    ):
        self._hosts = hosts
        self._user = user
        self._password = password
        self._vhost = vhost
        self._publisher_confirms = publisher_confirms
        self._channel_qos_kwargs = channel_qos_kwargs or {}
        self._use_transaction = use_transaction
        self._use_ssl = use_ssl
        self._port = port or (DEFAULT_PORTS["amqps"] if use_ssl else DEFAULT_PORTS["amqp"])

        self._connection: Optional[AbstractRobustConnection] = None
        self._channel: Optional[AbstractRobustChannel] = None

    async def connect(self) -> None:
        if self._connection is None or self._connection.is_closed or \
                self._channel is None or self._channel.is_closed:
            await self._reconnect()

    async def close(self) -> None:
        await self._close_channel()
        await self._close_connection()

    @property
    async def connection(self) -> AbstractRobustConnection:
        if self._connection is None or self._connection.is_closed:
            await self._reconnect()
        return self._connection

    @property
    async def channel(self) -> AbstractRobustChannel:
        if self._channel is None or self._channel.is_closed:
            await self._reconnect()
        return self._channel

    async def _reconnect(self) -> None:
        if self._connection is None or self._connection.is_closed:
            await self._create_connection()
        if self._channel is None or self._channel.is_closed:
            await self._create_channel()

    async def _create_connection(self) -> None:
        self._connection = await connect_robust(
            make_url(
                host=random.choice(self._hosts),
                port=self._port,
                login=self._user,
                password=self._password,
                virtualhost=self._vhost,
                ssl=self._use_ssl,
            ),
        )

    async def _create_channel(self) -> None:
        self._channel = await self._connection.channel(
            self._publisher_confirms,
        )
        await self._channel.set_qos(**self._channel_qos_kwargs)

    async def _close_connection(self) -> None:
        with NoException():
            await self._connection.close()

    async def _close_channel(self) -> None:
        with NoException():
            await self._channel.close()
